@@ -787,8 +787,8 @@ class xtables(object):
         self.proto = proto
         self._xt_globals = xtables_globals()
         self._xt_globals.option_offset = 0
-        self._xt_globals.program_name = version.__pkgname__
-        self._xt_globals.program_version = version.__version__
+        self._xt_globals.program_name = version.__pkgname__.encode()
+        self._xt_globals.program_version = version.__version__.encode()
         self._xt_globals.orig_opts = None
         self._xt_globals.opts = None
         self._xt_globals.exit_err = _xt_exit
@@ -846,8 +846,8 @@ class xtables(object):
         xtables._xtables_pending_targets.value = self._pending_targets
 
     def _check_extname(self, name):
-        if name in ["", "ACCEPT", "DROP", "QUEUE", "RETURN"]:
-            name = "standard"
+        if name in [b"", b"ACCEPT", b"DROP", b"QUEUE", b"RETURN"]:
+            name = b"standard"
         return name
 
     def _loaded(self, name):
@@ -888,6 +888,8 @@ class xtables(object):
             raise XTablesError("Unknown protocol %d" % (self.proto))
 
     def _try_register(self, name):
+        if isinstance(name, bytes):
+            name = name.decode()
         if self._try_extinit(name, _lib_xtables):
             return
         prefix = self._get_prefix()
@@ -899,6 +901,8 @@ class xtables(object):
 
     @preserve_globals
     def find_match(self, name):
+        if isinstance(name, str):
+            name = name.encode()
         name = self._check_extname(name)
         match = xtables._xtables_find_match(name, XTF_TRY_LOAD, None)
         if not match:
@@ -912,6 +916,8 @@ class xtables(object):
 
     @preserve_globals
     def find_target(self, name):
+        if isinstance(name, str):
+            name = name.encode()
         name = self._check_extname(name)
         target = xtables._xtables_find_target(name, XTF_TRY_LOAD)
         if not target:
